"""
Author: Wenyu Ouyang
Date: 2024-04-08 18:15:48
LastEditTime: 2024-09-16 10:19:34
LastEditors: Wenyu Ouyang
Description: HydroDL model class
FilePath: \torchhydro\torchhydro\trainers\deep_hydro.py
Copyright (c) 2024-2024 Wenyu Ouyang. All rights reserved.
"""

import copy
import os
from abc import ABC, abstractmethod
from collections import defaultdict
from functools import reduce
from typing import Dict, Tuple

import numpy as np
import torch
import torch.distributed as dist
import torch.nn as nn
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.optim.lr_scheduler import *
from torch.utils.data import DataLoader, DistributedSampler
from tqdm import tqdm

from torchhydro.configs.config import update_nested_dict
from torchhydro.datasets.data_dict import datasets_dict
from torchhydro.datasets.data_sets import BaseDataset
from torchhydro.datasets.sampler import (
    KuaiSampler,
    fl_sample_basin,
    fl_sample_region,
    HydroSampler,
)
from torchhydro.models.model_dict_function import (
    pytorch_criterion_dict,
    pytorch_model_dict,
    pytorch_opt_dict,
)
from torchhydro.models.model_utils import get_the_device
from torchhydro.trainers.train_logger import TrainLogger
from torchhydro.trainers.train_utils import (
    EarlyStopper,
    average_weights,
    denormalize4eval,
    evaluate_validation,
    compute_validation,
    model_infer,
    read_pth_from_model_loader,
    torch_single_train,
    calculate_and_record_metrics,
)


class DeepHydroInterface(ABC):
    """
    An abstract class used to handle different configurations
    of hydrological deep learning models + hyperparams for training, test, and predict functions.
    This class assumes that data is already split into test train and validation at this point.
    """

    def __init__(self, cfgs: Dict):
        """
        Parameters
        ----------
        cfgs
            configs for initializing DeepHydro
        """

        self._cfgs = cfgs

    @property
    def cfgs(self):
        """all configs"""
        return self._cfgs

    @property
    def weight_path(self):
        """weight path"""
        return self._cfgs["model_cfgs"]["weight_path"]

    @weight_path.setter
    def weight_path(self, weight_path):
        self._cfgs["model_cfgs"]["weight_path"] = weight_path

    @abstractmethod
    def load_model(self, mode="train") -> object:
        """Get a Hydro DL model"""
        raise NotImplementedError

    @abstractmethod
    def make_dataset(self, is_tra_val_te: str) -> object:
        """
        Initializes a pytorch dataset.

        Parameters
        ----------
        is_tra_val_te
            train or valid or test

        Returns
        -------
        object
            a dataset class loading data from data source
        """
        raise NotImplementedError

    @abstractmethod
    def model_train(self):
        """
        Train the model
        """
        raise NotImplementedError

    @abstractmethod
    def model_evaluate(self):
        """
        Evaluate the model
        """
        raise NotImplementedError


class DeepHydro(DeepHydroInterface):
    """
    The Base Trainer class for Hydrological Deep Learning models
    """

    def __init__(
        self,
        cfgs: Dict,
        pre_model=None,
    ):
        """
        Parameters
        ----------
        cfgs
            configs for the model
        pre_model
            a pre-trained model, if it is not None,
            we will use its weights to initialize this model
            by default None
        """
        super().__init__(cfgs)
        self.device_num = cfgs["training_cfgs"]["device"]
        self.device = get_the_device(self.device_num)
        self.pre_model = pre_model
        self.model = self.load_model()
        if cfgs["training_cfgs"]["train_mode"]:
            self.traindataset = self.make_dataset("train")
            if cfgs["data_cfgs"]["t_range_valid"] is not None:
                self.validdataset = self.make_dataset("valid")
        self.testdataset: BaseDataset = self.make_dataset("test")
        print(f"Torch is using {str(self.device)}")

    def load_model(self, mode="train"):
        """
        Load a time series forecast model in pytorch_model_dict in model_dict_function.py

        Returns
        -------
        object
            model in pytorch_model_dict in model_dict_function.py
        """
        if mode == "infer":
            self.weight_path = self._get_trained_model()
        elif mode != "train":
            raise ValueError("Invalid mode; must be 'train' or 'infer'")
        model_cfgs = self.cfgs["model_cfgs"]
        model_name = model_cfgs["model_name"]
        if model_name not in pytorch_model_dict:
            raise NotImplementedError(
                f"Error the model {model_name} was not found in the model dict. Please add it."
            )
        if self.pre_model is not None:
            model = self._load_pretrain_model()
        elif self.weight_path is not None:
            # load model from pth file (saved weights and biases)
            model = self._load_model_from_pth()
        else:
            model = pytorch_model_dict[model_name](**model_cfgs["model_hyperparam"])
            # model_data = torch.load(weight_path)
            # model.load_state_dict(model_data)
        if torch.cuda.device_count() > 1 and len(self.device_num) > 1:
            print("Let's use", torch.cuda.device_count(), "GPUs!")
            which_first_tensor = self.cfgs["training_cfgs"]["which_first_tensor"]
            sequece_first = which_first_tensor == "sequence"
            parallel_dim = 1 if sequece_first else 0
            model = nn.DataParallel(model, device_ids=self.device_num, dim=parallel_dim)
        model.to(self.device)
        return model

    def _load_pretrain_model(self):
        """load a pretrained model as the initial model"""
        return self.pre_model

    def _load_model_from_pth(self):
        weight_path = self.weight_path
        model_cfgs = self.cfgs["model_cfgs"]
        model_name = model_cfgs["model_name"]
        model = pytorch_model_dict[model_name](**model_cfgs["model_hyperparam"])
        checkpoint = torch.load(weight_path, map_location=self.device)
        model.load_state_dict(checkpoint)
        print("Weights sucessfully loaded")
        return model

    def make_dataset(self, is_tra_val_te: str):
        """
        Initializes a pytorch dataset.

        Parameters
        ----------
        is_tra_val_te
            train or valid or test

        Returns
        -------
        object
            an object initializing from class in datasets_dict in data_dict.py
        """
        data_cfgs = self.cfgs["data_cfgs"]
        dataset_name = data_cfgs["dataset"]

        if dataset_name in list(datasets_dict.keys()):
            dataset = datasets_dict[dataset_name](data_cfgs, is_tra_val_te)
        else:
            raise NotImplementedError(
                f"Error the dataset {str(dataset_name)} was not found in the dataset dict. Please add it."
            )
        return dataset

    def model_train(self) -> None:
        """train a hydrological DL model"""
        # A dictionary of the necessary parameters for training
        training_cfgs = self.cfgs["training_cfgs"]
        # The file path to load model weights from; defaults to "model_save"
        model_filepath = self.cfgs["data_cfgs"]["test_path"]
        data_cfgs = self.cfgs["data_cfgs"]
        es = None
        if training_cfgs["early_stopping"]:
            es = EarlyStopper(training_cfgs["patience"])
        criterion = self._get_loss_func(training_cfgs)
        opt = self._get_optimizer(training_cfgs)
        scheduler = self._get_scheduler(training_cfgs, opt)
        max_epochs = training_cfgs["epochs"]
        start_epoch = training_cfgs["start_epoch"]
        # use PyTorch's DataLoader to load the data into batches in each epoch
        data_loader, validation_data_loader = self._get_dataloader(
            training_cfgs, data_cfgs
        )
        logger = TrainLogger(model_filepath, self.cfgs, opt)
        for epoch in range(start_epoch, max_epochs + 1):
            with logger.log_epoch_train(epoch) as train_logs:
                total_loss, n_iter_ep = torch_single_train(
                    self.model,
                    opt,
                    criterion,
                    data_loader,
                    device=self.device,
                    which_first_tensor=training_cfgs["which_first_tensor"],
                )
                train_logs["train_loss"] = total_loss
                train_logs["model"] = self.model

            valid_loss = None
            valid_metrics = None
            if data_cfgs["t_range_valid"] is not None:
                with logger.log_epoch_valid(epoch) as valid_logs:
                    valid_loss, valid_metrics = self._1epoch_valid(
                        training_cfgs, criterion, validation_data_loader, valid_logs
                    )

            self._scheduler_step(training_cfgs, scheduler, valid_loss)
            logger.save_session_param(
                epoch, total_loss, n_iter_ep, valid_loss, valid_metrics
            )
            logger.save_model_and_params(self.model, epoch, self.cfgs)
            if es and not es.check_loss(
                self.model,
                valid_loss,
                self.cfgs["data_cfgs"]["test_path"],
            ):
                print("Stopping model now")
                break
        # logger.plot_model_structure(self.model)
        logger.tb.close()

        # return the trained model weights and bias and the epoch loss
        return self.model.state_dict(), sum(logger.epoch_loss) / len(logger.epoch_loss)

    def _get_scheduler(self, training_cfgs, opt):
        lr_scheduler_cfg = training_cfgs["lr_scheduler"]

        if "lr" in lr_scheduler_cfg and "lr_factor" not in lr_scheduler_cfg:
            scheduler = LambdaLR(opt, lr_lambda=lambda epoch: 1.0)
        elif isinstance(lr_scheduler_cfg, dict) and all(
            isinstance(epoch, int) for epoch in lr_scheduler_cfg
        ):
            scheduler = LambdaLR(
                opt, lr_lambda=lambda epoch: lr_scheduler_cfg.get(epoch, 1.0)
            )
        elif "lr_factor" in lr_scheduler_cfg and "lr_patience" not in lr_scheduler_cfg:
            scheduler = ExponentialLR(opt, gamma=lr_scheduler_cfg["lr_factor"])
        elif "lr_factor" in lr_scheduler_cfg:
            scheduler = ReduceLROnPlateau(
                opt,
                mode="min",
                factor=lr_scheduler_cfg["lr_factor"],
                patience=lr_scheduler_cfg["lr_patience"],
            )
        else:
            raise ValueError("Invalid lr_scheduler configuration")

        return scheduler

    def _scheduler_step(self, training_cfgs, scheduler, valid_loss):
        lr_scheduler_cfg = training_cfgs["lr_scheduler"]
        required_keys = {"lr_factor", "lr_patience"}
        if required_keys.issubset(lr_scheduler_cfg.keys()):
            scheduler.step(valid_loss)
        else:
            scheduler.step()

    def _1epoch_valid(
        self, training_cfgs, criterion, validation_data_loader, valid_logs
    ):
        valid_obss_np, valid_preds_np, valid_loss = compute_validation(
            self.model,
            criterion,
            validation_data_loader,
            device=self.device,
            which_first_tensor=training_cfgs["which_first_tensor"],
        )
        valid_logs["valid_loss"] = valid_loss
        if self.cfgs["evaluation_cfgs"]["calc_metrics"]:
            target_col = self.cfgs["data_cfgs"]["target_cols"]
            valid_metrics = evaluate_validation(
                validation_data_loader,
                valid_preds_np,
                valid_obss_np,
                self.cfgs["evaluation_cfgs"],
                target_col,
            )
            valid_logs["valid_metrics"] = valid_metrics
            return valid_loss, valid_metrics
        return valid_loss, None

    def _get_trained_model(self):
        model_loader = self.cfgs["evaluation_cfgs"]["model_loader"]
        model_pth_dir = self.cfgs["data_cfgs"]["test_path"]
        return read_pth_from_model_loader(model_loader, model_pth_dir)

    def model_evaluate(self) -> Tuple[Dict, np.array, np.array]:
        """
        A function to evaluate a model, called at end of training.

        Returns
        -------
        tuple[dict, np.array, np.array]
            eval_log, denormalized predictions and observations
        """
        self.model = self.load_model(mode="infer")
        preds_xr, obss_xr = self.inference()
        return preds_xr, obss_xr

    def inference(self) -> Tuple[torch.Tensor, torch.Tensor]:
        """infer using trained model and unnormalized results"""
        data_cfgs = self.cfgs["data_cfgs"]
        training_cfgs = self.cfgs["training_cfgs"]
        evaluation_cfgs = self.cfgs["evaluation_cfgs"]
        device = get_the_device(self.cfgs["training_cfgs"]["device"])

        ngrid = self.testdataset.ngrid
        if data_cfgs["sampler"] == "HydroSampler":
            test_num_samples = self.testdataset.num_samples
            test_dataloader = DataLoader(
                self.testdataset,
                batch_size=test_num_samples // ngrid,
                shuffle=False,
                drop_last=False,
                timeout=0,
            )
        else:
            test_dataloader = DataLoader(
                self.testdataset,
                batch_size=training_cfgs["batch_size"],
                shuffle=False,
                sampler=None,
                batch_sampler=None,
                drop_last=False,
                timeout=0,
                worker_init_fn=None,
            )
        seq_first = training_cfgs["which_first_tensor"] == "sequence"
        self.model.eval()
        # here the batch is just an index of lookup table, so any batch size could be chosen
        test_preds = []
        obss = []
        with torch.no_grad():
            for xs, ys in test_dataloader:
                # here the a batch doesn't mean a basin; it is only an index in lookup table
                # for NtoN mode, only basin is index in lookup table, so the batch is same as basin
                # for Nto1 mode, batch is only an index
                ys, pred = model_infer(seq_first, device, self.model, xs, ys)
                test_preds.append(pred.cpu().numpy())
                obss.append(ys.cpu().numpy())
            pred = reduce(lambda x, y: np.vstack((x, y)), test_preds)
            obs = reduce(lambda x, y: np.vstack((x, y)), obss)
        if pred.ndim == 2:
            # TODO: check
            # the ndim is 2 meaning we use an Nto1 mode
            # as lookup table is (basin 1's all time length, basin 2's all time length, ...)
            # params of reshape should be (basin size, time length)
            pred = pred.flatten().reshape(test_dataloader.test_data.y.shape[0], -1, 1)
            obs = obs.flatten().reshape(test_dataloader.test_data.y.shape[0], -1, 1)

        if not evaluation_cfgs["long_seq_pred"]:
            target_len = len(data_cfgs["target_cols"])
            prec_window = data_cfgs["prec_window"]
            batch_size = test_dataloader.batch_size
            if evaluation_cfgs["rolling"]:
                forecast_length = data_cfgs["forecast_length"]
                pred = pred[:, prec_window:, :].reshape(
                    ngrid, batch_size, forecast_length, target_len
                )
                obs = obs[:, prec_window:, :].reshape(
                    ngrid, batch_size, forecast_length, target_len
                )

                pred = pred[:, ::forecast_length, :, :]
                obs = obs[:, ::forecast_length, :, :]

                pred = np.concatenate(pred, axis=0).reshape(ngrid, -1, target_len)
                obs = np.concatenate(obs, axis=0).reshape(ngrid, -1, target_len)

                pred = pred[:, :batch_size, :]
                obs = obs[:, :batch_size, :]
            else:
                pred = pred[:, prec_window, :].reshape(ngrid, batch_size, target_len)
                obs = obs[:, prec_window, :].reshape(ngrid, batch_size, target_len)
            pred_xr, obs_xr = denormalize4eval(
                test_dataloader, pred, obs, long_seq_pred=False
            )
            fill_nan = evaluation_cfgs["fill_nan"]
            eval_log = {}
            for i, col in enumerate(data_cfgs["target_cols"]):
                obs = obs_xr[col].to_numpy()
                pred = pred_xr[col].to_numpy()
                eval_log = calculate_and_record_metrics(
                    obs,
                    pred,
                    evaluation_cfgs["metrics"],
                    col,
                    fill_nan[i] if isinstance(fill_nan, list) else fill_nan,
                    eval_log,
                )
            test_log = f" Best Metric {eval_log}"
            print(test_log)
        else:
            pred_xr, obs_xr = denormalize4eval(test_dataloader, pred, obs)
        return pred_xr, obs_xr

    def _get_optimizer(self, training_cfgs):
        params_in_opt = self.model.parameters()
        return pytorch_opt_dict[training_cfgs["optimizer"]](
            params_in_opt, **training_cfgs["optim_params"]
        )

    def _get_loss_func(self, training_cfgs):
        criterion_init_params = {}
        if "criterion_params" in training_cfgs:
            loss_param = training_cfgs["criterion_params"]
            if loss_param is not None:
                for key in loss_param.keys():
                    if key == "loss_funcs":
                        criterion_init_params[key] = pytorch_criterion_dict[
                            loss_param[key]
                        ]()
                    else:
                        criterion_init_params[key] = loss_param[key]
        return pytorch_criterion_dict[training_cfgs["criterion"]](
            **criterion_init_params
        )

    def _get_dataloader(self, training_cfgs, data_cfgs):
        worker_num = 0
        pin_memory = False
        if "num_workers" in training_cfgs:
            worker_num = training_cfgs["num_workers"]
            print(f"using {str(worker_num)} workers")
        if "pin_memory" in training_cfgs:
            pin_memory = training_cfgs["pin_memory"]
            print(f"Pin memory set to {str(pin_memory)}")
        train_dataset: BaseDataset = self.traindataset
        sampler = None
        if data_cfgs["sampler"] is not None:
            # now we only have one special sampler from Kuai Fang's Deep Learning papers
            batch_size = data_cfgs["batch_size"]
            rho = data_cfgs["forecast_history"]
            warmup_length = data_cfgs["warmup_length"]
            horizon = data_cfgs["forecast_length"]
            ngrid = train_dataset.ngrid
            nt = train_dataset.nt
            if data_cfgs["sampler"] == "HydroSampler":
                sampler = HydroSampler(train_dataset)
            elif data_cfgs["sampler"] == "KuaiSampler":
                sampler = KuaiSampler(
                    train_dataset,
                    batch_size=batch_size,
                    warmup_length=warmup_length,
                    rho_horizon=rho + horizon,
                    ngrid=ngrid,
                    nt=nt,
                )
            elif data_cfgs["sampler"] == "DistSampler":
                sampler = DistributedSampler(train_dataset)
            else:
                raise NotImplementedError("This sampler not implemented yet")
        data_loader = DataLoader(
            train_dataset,
            batch_size=training_cfgs["batch_size"],
            shuffle=(sampler is None),
            sampler=sampler,
            num_workers=worker_num,
            pin_memory=pin_memory,
            timeout=0,
        )
        if data_cfgs["t_range_valid"] is not None:
            valid_dataset: BaseDataset = self.validdataset
            batch_size_valid = training_cfgs["batch_size"]
            if data_cfgs["sampler"] == "HydroSampler":
                # for HydroSampler when evaluating, we need to set new batch size
                batch_size_valid = valid_dataset.num_samples // ngrid
            validation_data_loader = DataLoader(
                valid_dataset,
                batch_size=batch_size_valid,
                shuffle=False,
                num_workers=worker_num,
                pin_memory=pin_memory,
                timeout=0,
            )
            return data_loader, validation_data_loader

        return data_loader, None


class FedLearnHydro(DeepHydro):
    """Federated Learning Hydrological DL model"""

    def __init__(self, cfgs: Dict):
        super().__init__(cfgs)
        # a user group which is a dict where the keys are the user index
        # and the values are the corresponding data for each of those users
        train_dataset = self.traindataset
        fl_hyperparam = self.cfgs["model_cfgs"]["fl_hyperparam"]
        # sample training data amongst users
        if fl_hyperparam["fl_sample"] == "basin":
            # Sample a basin for a user
            user_groups = fl_sample_basin(train_dataset)
        elif fl_hyperparam["fl_sample"] == "region":
            # Sample a region for a user
            user_groups = fl_sample_region(train_dataset)
        else:
            raise NotImplementedError()
        self.user_groups = user_groups

    @property
    def num_users(self):
        """number of users in federated learning"""
        return len(self.user_groups)

    def model_train(self) -> None:
        # BUILD MODEL
        global_model = self.model

        # copy weights
        global_weights = global_model.state_dict()

        # Training
        train_loss, train_accuracy = [], []
        print_every = 2

        training_cfgs = self.cfgs["training_cfgs"]
        model_cfgs = self.cfgs["model_cfgs"]
        max_epochs = training_cfgs["epochs"]
        start_epoch = training_cfgs["start_epoch"]
        fl_hyperparam = model_cfgs["fl_hyperparam"]
        # total rounds in a FL system is max_epochs
        for epoch in tqdm(range(start_epoch, max_epochs + 1)):
            local_weights, local_losses = [], []
            print(f"\n | Global Training Round : {epoch} |\n")

            global_model.train()
            m = max(int(fl_hyperparam["fl_frac"] * self.num_users), 1)
            # randomly select m users, they will be the clients in this round
            idxs_users = np.random.choice(range(self.num_users), m, replace=False)

            for idx in idxs_users:
                # each user will be used to train the model locally
                # user_gourps[idx] means the idx of dataset for a user
                user_cfgs = self._get_a_user_cfgs(idx)
                local_model = DeepHydro(
                    user_cfgs,
                    pre_model=copy.deepcopy(global_model),
                )
                w, loss = local_model.model_train()
                local_weights.append(copy.deepcopy(w))
                local_losses.append(copy.deepcopy(loss))

            # update global weights
            global_weights = average_weights(local_weights)

            # update global weights
            global_model.load_state_dict(global_weights)

            loss_avg = sum(local_losses) / len(local_losses)
            train_loss.append(loss_avg)

            # Calculate avg training accuracy over all users at every epoch
            list_acc = []
            global_model.eval()
            for c in range(self.num_users):
                one_user_cfg = self._get_a_user_cfgs(c)
                local_model = DeepHydro(
                    one_user_cfg,
                    pre_model=global_model,
                )
                acc, _, _ = local_model.model_evaluate()
                list_acc.append(acc)
            values = [list(d.values())[0][0] for d in list_acc]
            filtered_values = [v for v in values if not np.isnan(v)]
            train_accuracy.append(sum(filtered_values) / len(filtered_values))

            # print global training loss after every 'i' rounds
            if (epoch + 1) % print_every == 0:
                print(f" \nAvg Training Stats after {epoch+1} global rounds:")
                print(f"Training Loss : {np.mean(np.array(train_loss))}")
                print("Train Accuracy: {:.2f}% \n".format(100 * train_accuracy[-1]))

    def _get_a_user_cfgs(self, idx):
        """To get a user's configs for local training"""
        user = self.user_groups[idx]

        # update data_cfgs
        # Use defaultdict to collect dates for each basin
        basin_dates = defaultdict(list)

        for _, (basin, time) in user.items():
            basin_dates[basin].append(time)

        # Initialize a list to store distinct basins
        basins = []

        # for each basin, we can find its date range
        date_ranges = {}
        for basin, times in basin_dates.items():
            basins.append(basin)
            date_ranges[basin] = (np.min(times), np.max(times))
        # get the longest date range
        longest_date_range = max(date_ranges.values(), key=lambda x: x[1] - x[0])
        # transform the date range of numpy data into string
        longest_date_range = [
            np.datetime_as_string(dt, unit="D") for dt in longest_date_range
        ]
        user_cfgs = copy.deepcopy(self.cfgs)
        # update data_cfgs
        update_nested_dict(
            user_cfgs, ["data_cfgs", "t_range_train"], longest_date_range
        )
        # for local training in FL, we don't need a validation set
        update_nested_dict(user_cfgs, ["data_cfgs", "t_range_valid"], None)
        # for local training in FL, we don't need a test set, but we should set one to avoid error
        update_nested_dict(user_cfgs, ["data_cfgs", "t_range_test"], longest_date_range)
        update_nested_dict(user_cfgs, ["data_cfgs", "object_ids"], basins)

        # update training_cfgs
        # we also need to update some training params for local training from FL settings
        update_nested_dict(
            user_cfgs,
            ["training_cfgs", "epochs"],
            user_cfgs["model_cfgs"]["fl_hyperparam"]["fl_local_ep"],
        )
        update_nested_dict(
            user_cfgs,
            ["evaluation_cfgs", "test_epoch"],
            user_cfgs["model_cfgs"]["fl_hyperparam"]["fl_local_ep"],
        )
        # don't need to save model weights for local training
        update_nested_dict(
            user_cfgs,
            ["training_cfgs", "save_epoch"],
            None,
        )
        # there are two settings for batch size in configs, we need to update both of them
        update_nested_dict(
            user_cfgs,
            ["training_cfgs", "batch_size"],
            user_cfgs["model_cfgs"]["fl_hyperparam"]["fl_local_bs"],
        )
        update_nested_dict(
            user_cfgs,
            ["data_cfgs", "batch_size"],
            user_cfgs["model_cfgs"]["fl_hyperparam"]["fl_local_bs"],
        )

        # update model_cfgs finally
        # For local model, its model_type is Normal
        update_nested_dict(user_cfgs, ["model_cfgs", "model_type"], "Normal")
        update_nested_dict(
            user_cfgs,
            ["model_cfgs", "fl_hyperparam"],
            None,
        )
        return user_cfgs


class TransLearnHydro(DeepHydro):
    def __init__(self, cfgs: Dict, pre_model=None):
        super().__init__(cfgs, pre_model)

    def load_model(self, mode="train"):
        """Load model for transfer learning"""
        model_cfgs = self.cfgs["model_cfgs"]
        if self.weight_path is None and self.pre_model is None:
            raise NotImplementedError(
                "For transfer learning, we need a pre-trained model"
            )
        model = super().load_model(mode)
        if (
            "weight_path_add" in model_cfgs
            and "freeze_params" in model_cfgs["weight_path_add"]
        ):
            freeze_params = model_cfgs["weight_path_add"]["freeze_params"]
            for param in freeze_params:
                exec(f"model.{param}.requires_grad = False")
        return model

    def _load_model_from_pth(self):
        weight_path = self.weight_path
        model_cfgs = self.cfgs["model_cfgs"]
        model_name = model_cfgs["model_name"]
        model = pytorch_model_dict[model_name](**model_cfgs["model_hyperparam"])
        checkpoint = torch.load(weight_path, map_location=self.device)
        if "weight_path_add" in model_cfgs:
            if "excluded_layers" in model_cfgs["weight_path_add"]:
                # delete some layers from source model if we don't need them
                excluded_layers = model_cfgs["weight_path_add"]["excluded_layers"]
                for layer in excluded_layers:
                    del checkpoint[layer]
                print("sucessfully deleted layers")
            else:
                print("directly loading identically-named layers of source model")
        model.load_state_dict(checkpoint, strict=False)
        print("Weights sucessfully loaded")
        return model


class MultiTaskHydro(DeepHydro):
    def __init__(self, cfgs: Dict, pre_model=None):
        super().__init__(cfgs, pre_model)

    def _get_optimizer(self, training_cfgs):
        params_in_opt = self.model.parameters()
        if training_cfgs["criterion"] == "UncertaintyWeights":
            # log_var = torch.zeros((1,), requires_grad=True)
            log_vars = [
                torch.zeros((1,), requires_grad=True, device=self.device)
                for _ in range(training_cfgs["multi_targets"])
            ]
            params_in_opt = list(self.model.parameters()) + log_vars
        return pytorch_opt_dict[training_cfgs["optimizer"]](
            params_in_opt, **training_cfgs["optim_params"]
        )

    def _get_loss_func(self, training_cfgs):
        if "criterion_params" in training_cfgs:
            loss_param = training_cfgs["criterion_params"]
            if loss_param is not None:
                criterion_init_params = {
                    key: (
                        pytorch_criterion_dict[loss_param[key]]()
                        if key == "loss_funcs"
                        else loss_param[key]
                    )
                    for key in loss_param.keys()
                }
        if training_cfgs["criterion"] == "MultiOutWaterBalanceLoss":
            # TODO: hard code for streamflow and ET
            stat_dict = self.traindataset.target_scaler.stat_dict
            stat_dict_keys = list(stat_dict.keys())
            q_name = np.intersect1d(
                [
                    "usgsFlow",
                    "streamflow",
                    "Q",
                    "qobs",
                ],
                stat_dict_keys,
            )[0]
            et_name = np.intersect1d(
                [
                    "ET",
                    "LE",
                    "GPP",
                    "Ec",
                    "Es",
                    "Ei",
                    "ET_water",
                    # sum pf ET components in PML V2
                    "ET_sum",
                ],
                stat_dict_keys,
            )[0]
            q_mean = self.training.target_scaler.stat_dict[q_name][2]
            q_std = self.training.target_scaler.stat_dict[q_name][3]
            et_mean = self.training.target_scaler.stat_dict[et_name][2]
            et_std = self.training.target_scaler.stat_dict[et_name][3]
            means = [q_mean, et_mean]
            stds = [q_std, et_std]
            criterion_init_params["means"] = means
            criterion_init_params["stds"] = stds
        return pytorch_criterion_dict[training_cfgs["criterion"]](
            **criterion_init_params
        )


class DistributedDeepHydro(MultiTaskHydro):
    def __init__(self, world_size, cfgs: Dict):
        super().__init__(cfgs, cfgs["model_cfgs"]["weight_path"])
        self.world_size = world_size

    def setup(self, rank):
        os.environ["MASTER_ADDR"] = self.cfgs["training_cfgs"]["master_addr"]
        os.environ["MASTER_PORT"] = self.cfgs["training_cfgs"]["port"]
        dist.init_process_group(
            backend="nccl", init_method="env://", world_size=self.world_size, rank=rank
        )
        torch.cuda.set_device(rank)
        self.device = torch.device(rank)
        self.rank = rank

    def cleanup(self):
        dist.destroy_process_group()

    def load_model(self, mode="train"):
        if mode == "infer":
            if self.weight_path is None:
                # for continue training
                self.weight_path = self._get_trained_model()
        elif mode != "train":
            raise ValueError("Invalid mode; must be 'train' or 'infer'")
        model_cfgs = self.cfgs["model_cfgs"]
        model_name = model_cfgs["model_name"]
        if model_name not in pytorch_model_dict:
            raise NotImplementedError(
                f"Error the model {model_name} was not found in the model dict. Please add it."
            )
        if self.pre_model is not None:
            return self._load_pretrain_model()
        elif self.weight_path is not None:
            # load model from pth file (saved weights and biases)
            return self._load_model_from_pth()
        else:
            return pytorch_model_dict[model_name](**model_cfgs["model_hyperparam"])

    def model_train(self):
        model = self.load_model().to(self.device)
        self.model = DDP(model, device_ids=[self.rank])
        training_cfgs = self.cfgs["training_cfgs"]
        # The file path to load model weights from; defaults to "model_save"
        model_filepath = self.cfgs["data_cfgs"]["test_path"]
        data_cfgs = self.cfgs["data_cfgs"]
        es = None
        if training_cfgs["early_stopping"]:
            es = EarlyStopper(training_cfgs["patience"])
        criterion = self._get_loss_func(training_cfgs)
        opt = self._get_optimizer(training_cfgs)
        scheduler = self._get_scheduler(training_cfgs, opt)
        max_epochs = training_cfgs["epochs"]
        start_epoch = training_cfgs["start_epoch"]
        # use PyTorch's DataLoader to load the data into batches in each epoch
        data_loader, validation_data_loader = self._get_dataloader(
            training_cfgs, data_cfgs
        )
        logger = TrainLogger(model_filepath, self.cfgs, opt)
        for epoch in range(start_epoch, max_epochs + 1):
            data_loader.sampler.set_epoch(epoch)
            with logger.log_epoch_train(epoch) as train_logs:
                total_loss, n_iter_ep = torch_single_train(
                    self.model,
                    opt,
                    criterion,
                    data_loader,
                    device=self.device,
                    which_first_tensor=training_cfgs["which_first_tensor"],
                )
                train_logs["train_loss"] = total_loss
                train_logs["model"] = self.model

            valid_loss = None
            valid_metrics = None
            if data_cfgs["t_range_valid"] is not None:
                with logger.log_epoch_valid(epoch) as valid_logs:
                    valid_loss, valid_metrics = self._1epoch_valid(
                        training_cfgs, criterion, validation_data_loader, valid_logs
                    )

            self._scheduler_step(training_cfgs, scheduler, valid_loss)
            logger.save_session_param(
                epoch, total_loss, n_iter_ep, valid_loss, valid_metrics
            )
            logger.save_model_and_params(self.model, epoch, self.cfgs)
            if es and not es.check_loss(
                self.model,
                valid_loss,
                self.cfgs["data_cfgs"]["test_path"],
            ):
                print("Stopping model now")
                break
        # if self.rank == 0:
        # logging.log(1, f"Training complete in: {time.time() - start_time:.2f} seconds"

    def run(self):
        self.setup(self.rank)
        self.model_train()
        self.model_evaluate()
        self.cleanup()


def train_worker(rank, world_size, cfgs):
    trainer = DistributedDeepHydro(world_size, cfgs)
    trainer.rank = rank
    trainer.run()


model_type_dict = {
    "Normal": DeepHydro,
    "FedLearn": FedLearnHydro,
    "TransLearn": TransLearnHydro,
    "MTL": MultiTaskHydro,
    "DDP_MTL": DistributedDeepHydro,
}



"""
Author: Wenyu Ouyang
Date: 2023-09-21 15:37:58
LastEditTime: 2024-07-17 19:13:41
LastEditors: Wenyu Ouyang
Description: Some basic funtions for dealing with data
FilePath: \torchhydro\torchhydro\datasets\data_utils.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

from typing import Union
from collections import OrderedDict
import numpy as np
import xarray as xr
import pint_xarray  # noqa: F401
import warnings


def warn_if_nan(dataarray, max_display=5, nan_mode="any"):
    """
    Issue a warning if the dataarray contains any NaN values and display their locations.

    Parameters
    -----------
    dataarray: xr.DataArray
        Input dataarray to check for NaN values.
    max_display: int
        Maximum number of NaN locations to display in the warning.
    nan_mode: str
        Mode of NaN checking: 'any' for any NaNs, 'all' for all values being NaNs.
    """
    if dataarray is None:
        return
    if nan_mode not in ["any", "all"]:
        raise ValueError("nan_mode must be 'any' or 'all'")

    if nan_mode == "all" and np.all(np.isnan(dataarray.values)):
        raise ValueError("The dataarray contains only NaN values!")

    nan_indices = np.argwhere(np.isnan(dataarray.values))
    total_nans = len(nan_indices)

    if total_nans <= 0:
        return False
    message = f"The dataarray contains {total_nans} NaN values!"

    # Displaying only the first few NaN locations if there are too many
    display_indices = nan_indices[:max_display].tolist()
    message += (
        f" Here are the indices of the first {max_display} NaNs: {display_indices}..."
        if total_nans > max_display
        else f" Here are the indices of the NaNs: {display_indices}"
    )
    warnings.warn(message)

    return True


def unify_streamflow_unit(ds: xr.Dataset, area=None, inverse=False):
    """Unify the unit of xr_dataset to be mm/day in a basin or inverse

    Parameters
    ----------
    ds: xarray dataset
        _description_
    area:
        area of each basin

    Returns
    -------
    _type_
        _description_
    """
    # use pint to convert unit
    if not inverse:
        target_unit = "mm/d"
        q = ds.pint.quantify()
        a = area.pint.quantify()
        r = q[list(q.keys())[0]] / a[list(a.keys())[0]]
        result = r.pint.to(target_unit).to_dataset(name=list(q.keys())[0])
    else:
        target_unit = "m^3/s"
        r = ds.pint.quantify()
        a = area.pint.quantify()
        q = r[list(r.keys())[0]] * a[list(a.keys())[0]]
        # q = q.pint.quantify()
        result = q.pint.to(target_unit).to_dataset(name=list(r.keys())[0])
    # dequantify to get normal xr_dataset
    return result.pint.dequantify()


def wrap_t_s_dict(data_cfgs: dict, is_tra_val_te: str) -> OrderedDict:
    """
    Basins and periods

    Parameters
    ----------

    data_cfgs
        configs for reading from data source
    is_tra_val_te
        train, valid or test

    Returns
    -------
    OrderedDict
        OrderedDict(sites_id=basins_id, t_final_range=t_range_list)
    """
    basins_id = data_cfgs["object_ids"]
    # if type(basins_id) is str and basins_id == "ALL":
    #     basins_id = data_source.read_object_ids().tolist()
    # assert all(x < y for x, y in zip(basins_id, basins_id[1:]))
    if f"t_range_{is_tra_val_te}" in data_cfgs:
        t_range_list = data_cfgs[f"t_range_{is_tra_val_te}"]
    else:
        raise Exception(
            f"Error! The mode {is_tra_val_te} was not found. Please add it."
        )
    return OrderedDict(sites_id=basins_id, t_final_range=t_range_list)


def _trans_norm(
    x: xr.DataArray,
    var_lst: list,
    stat_dict: dict,
    log_norm_cols: list = None,
    to_norm: bool = True,
    **kwargs,
) -> np.array:
    """
    Normalization or inverse normalization

    There are two normalization formulas:

    .. math:: normalized_x = (x - mean) / std

    and

     .. math:: normalized_x = [log_{10}(\sqrt{x} + 0.1) - mean] / std

     The later is only for vars in log_norm_cols; mean is mean value; std means standard deviation

    Parameters
    ----------
    x
        data to be normalized or denormalized
    var_lst
        the type of variables
    stat_dict
        statistics of all variables
    log_norm_cols
        which cols use the second norm method
    to_norm
        if true, normalize; else denormalize

    Returns
    -------
    np.array
        normalized or denormalized data
    """
    if x is None:
        return None
    if log_norm_cols is None:
        log_norm_cols = []
    if type(var_lst) is str:
        var_lst = [var_lst]
    out = xr.full_like(x, np.nan)
    for item in var_lst:
        stat = stat_dict[item]
        if to_norm:
            out.loc[dict(variable=item)] = (
                (np.log10(np.sqrt(np.abs(x.sel(variable=item))) + 0.1) - stat[2])
                / stat[3]
                if item in log_norm_cols
                else (x.sel(variable=item) - stat[2]) / stat[3]
            )
        elif item in log_norm_cols:
            out.loc[dict(variable=item)] = (
                np.power(10, x.sel(variable=item) * stat[3] + stat[2]) - 0.1
            ) ** 2
        else:
            out.loc[dict(variable=item)] = x.sel(variable=item) * stat[3] + stat[2]
    if to_norm:
        # after normalization, all units are dimensionless
        out.attrs = {}
    # after denormalization, recover units
    else:
        if "recover_units" in kwargs.keys() and kwargs["recover_units"] is not None:
            recover_units = kwargs["recover_units"]
            for item in var_lst:
                out.attrs["units"][item] = recover_units[item]
    return out


def _prcp_norm(x: np.array, mean_prep: np.array, to_norm: bool) -> np.array:
    """
    Normalize or denormalize data with mean precipitation.

    The formula is as follows when normalizing (denormalize equation is its inversion):

    .. math:: normalized_x = \frac{x}{precipitation}

    Parameters
    ----------
    x
        data to be normalized or denormalized
    mean_prep
        basins' mean precipitation
    to_norm
        if true, normalize; else denormalize

    Returns
    -------
    np.array
        normalized or denormalized data
    """
    tempprep = np.tile(mean_prep, (1, x.shape[1]))
    return x / tempprep if to_norm else x * tempprep


def dor_reservoirs_chosen(gages, usgs_id, dor_chosen) -> list:
    """
    choose basins of small DOR(calculated by NOR_STORAGE/RUNAVE7100)

    """

    dors = get_dor_values(gages, usgs_id)
    if type(dor_chosen) in [list, tuple]:
        # right half-open range
        chosen_id = [
            usgs_id[i]
            for i in range(dors.size)
            if dor_chosen[0] <= dors[i] < dor_chosen[1]
        ]
    elif dor_chosen < 0:
        chosen_id = [usgs_id[i] for i in range(dors.size) if dors[i] < -dor_chosen]
    else:
        chosen_id = [usgs_id[i] for i in range(dors.size) if dors[i] >= dor_chosen]

    assert all(x < y for x, y in zip(chosen_id, chosen_id[1:]))
    return chosen_id


def choose_sites_in_ecoregion(
    gages, site_ids: list, ecoregion: Union[list, tuple]
) -> list:
    """
    Choose sites in ecoregions

    Parameters
    ----------
    gages : Gages
        Only gages dataset has ecoregion attribute
    site_ids : list
        all ids of sites
    ecoregion : Union[list, tuple]
        which ecoregions

    Returns
    -------
    list
        chosen sites' ids

    Raises
    ------
    NotImplementedError
        PLease choose 'ECO2_CODE' or 'ECO3_CODE'
    NotImplementedError
        must be in EC02 code list
    NotImplementedError
        must be in EC03 code list
    """
    if ecoregion[0] not in ["ECO2_CODE", "ECO3_CODE"]:
        raise NotImplementedError("PLease choose 'ECO2_CODE' or 'ECO3_CODE'")
    if ecoregion[0] == "ECO2_CODE":
        ec02_code_lst = [
            5.2,
            5.3,
            6.2,
            7.1,
            8.1,
            8.2,
            8.3,
            8.4,
            8.5,
            9.2,
            9.3,
            9.4,
            9.5,
            9.6,
            10.1,
            10.2,
            10.4,
            11.1,
            12.1,
            13.1,
        ]
        if ecoregion[1] not in ec02_code_lst:
            raise NotImplementedError(
                f"No such EC02 code, please choose from {ec02_code_lst}"
            )
        attr_name = "ECO2_BAS_DOM"
    elif ecoregion[1] in np.arange(1, 85):
        attr_name = "ECO3_BAS_DOM"
    else:
        raise NotImplementedError("No such EC03 code, please choose from 1 - 85")
    attr_lst = [attr_name]
    data_attr = gages.read_constant_cols(site_ids, attr_lst)
    eco_names = data_attr[:, 0]
    return [site_ids[i] for i in range(eco_names.size) if eco_names[i] == ecoregion[1]]


def choose_basins_with_area(
    gages,
    usgs_ids: list,
    smallest_area: float,
    largest_area: float,
) -> list:
    """
    choose basins with not too large or too small area

    Parameters
    ----------
    gages
        Camels, CamelsSeries, Gages or GagesPro object
    usgs_ids: list
        given sites' ids
    smallest_area
        lower limit; unit is km2
    largest_area
        upper limit; unit is km2

    Returns
    -------
    list
        sites_chosen: [] -- ids of chosen gages

    """
    basins_areas = gages.read_basin_area(usgs_ids).flatten()
    sites_index = np.arange(len(usgs_ids))
    sites_chosen = np.ones(len(usgs_ids))
    for i in range(sites_index.size):
        # loop for every site
        if basins_areas[i] < smallest_area or basins_areas[i] > largest_area:
            sites_chosen[sites_index[i]] = 0
        else:
            sites_chosen[sites_index[i]] = 1
    return [usgs_ids[i] for i in range(len(sites_chosen)) if sites_chosen[i] > 0]


def diversion_chosen(gages, usgs_id):
    diversion_strs = ["diversion", "divert"]
    assert all(x < y for x, y in zip(usgs_id, usgs_id[1:]))
    attr_lst = ["WR_REPORT_REMARKS", "SCREENING_COMMENTS"]
    data_attr = gages.read_attr_origin(usgs_id, attr_lst)
    diversion_strs_lower = [elem.lower() for elem in diversion_strs]
    data_attr0_lower = np.array(
        [elem.lower() if type(elem) == str else elem for elem in data_attr[0]]
    )
    data_attr1_lower = np.array(
        [elem.lower() if type(elem) == str else elem for elem in data_attr[1]]
    )
    data_attr_lower = np.vstack((data_attr0_lower, data_attr1_lower)).T
    return [
        usgs_id[i]
        for i in range(len(usgs_id))
        if is_any_elem_in_a_lst(diversion_strs_lower, data_attr_lower[i], include=True)
    ]


def dam_num_chosen(gages, usgs_id, dam_num):
    """choose basins of dams"""
    assert all(x < y for x, y in zip(usgs_id, usgs_id[1:]))
    attr_lst = ["NDAMS_2009"]
    data_attr = gages.read_constant_cols(usgs_id, attr_lst)
    return (
        [
            usgs_id[i]
            for i in range(data_attr.size)
            if dam_num[0] <= data_attr[:, 0][i] < dam_num[1]
        ]
        if type(dam_num) == list
        else [
            usgs_id[i] for i in range(data_attr.size) if data_attr[:, 0][i] == dam_num
        ]
    )



"""
Author: Wenyu Ouyang
Date: 2023-07-25 16:47:19
LastEditTime: 2024-04-10 21:00:10
LastEditors: Wenyu Ouyang
Description: Test a full training and evaluating process
FilePath: \torchhydro\tests\test_train_camels_lstm.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

from torchhydro.configs.config import update_cfg
from torchhydro.trainers.trainer import train_and_evaluate


def test_train_evaluate(args, config_data):
    update_cfg(config_data, args)
    train_and_evaluate(config_data)



"""
Author: Wenyu Ouyang
Date: 2023-07-11 17:39:09
LastEditTime: 2023-10-06 18:12:54
LastEditors: Wenyu Ouyang
Description: Functions for dropout. Code is from Kuai Fang's repo: hydroDL
FilePath: \torchhydro\torchhydro\models\dropout.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""
import torch.nn


def create_mask(x, dr):
    """
    Dropout method in Gal & Ghahramami: A Theoretically Grounded Application of Dropout in RNNs.
    http://papers.nips.cc/paper/6241-a-theoretically-grounded-application-of-dropout-in-recurrent-neural-networks.pdf

    Parameters
    ----------
    x
    dr

    Returns
    -------

    """
    return x.new().resize_as_(x).bernoulli_(1 - dr).div_(1 - dr).detach_()


class DropMask(torch.autograd.function.InplaceFunction):
    @classmethod
    def forward(cls, ctx, input, mask, train=False, inplace=False):
        ctx.master_train = train
        ctx.inplace = inplace
        ctx.mask = mask

        if not ctx.master_train:
            return input
        if ctx.inplace:
            ctx.mark_dirty(input)
            output = input
        else:
            output = input.clone()
        output.mul_(ctx.mask)

        return output

    @staticmethod
    def backward(ctx, grad_output):
        if ctx.master_train:
            return grad_output * ctx.mask, None, None, None
        else:
            return grad_output, None, None, None



"""
Author: Wenyu Ouyang
Date: 2023-11-21 07:20:41
LastEditTime: 2023-12-18 09:16:10
LastEditors: Wenyu Ouyang
Description: Test weight analysis
FilePath: \torchhydro\tests\test_weight_analysis.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""
import os
import pandas as pd
import pytest


from torchhydro.explainers.weight_anlysis import (
    plot_param_hist_model,
)


@pytest.fixture()
def result_dir():
    project_dir = os.getcwd()
    return os.path.join(project_dir, "results")


def test_weight_analysis(result_dir):
    basin_id = "61561"
    # NOTICE: THE ORDER CANNOT BE MODIFIED WITHOUT DEBUGGING THE CODE IN plot_param_hist_model_fold
    chosen_layer_for_hist = [
        "former_linear.bias",
        "former_linear.weight",
        "linearIn.bias",
        "linearIn.weight",
        "linearOut.bias",
        "linearOut.weight",
        "lstm.b_hh",
        "lstm.b_ih",
        "lstm.w_hh",
        "lstm.w_ih",
    ]
    # too many figures lead to "Fail to allocate bitmap"
    show_hist_b = 20
    fold = 0
    exp61561_dir = os.path.join(result_dir, "test_camels", "expcccv61561_0")
    _, chosen_layers_consine = plot_param_hist_model(
        "lstm",
        exp61561_dir,
        show_hist_b,
        chosen_layer_for_hist,
        start_epoch=0,
        end_epoch=10,
        epoch_interval=1,
    )
    pd.DataFrame(chosen_layers_consine).to_csv(
        os.path.join(
            exp61561_dir,
            f"{basin_id}_bs{show_hist_b}_fold{fold}_chosen_layer_consine.csv",
        )
    )



import os
import pytest
from torchhydro.trainers.train_utils import read_pth_from_model_loader


def test_read_pth_from_model_loader_specified():
    model_loader = {"load_way": "specified", "test_epoch": 5}
    model_pth_dir = "/path/to/models"
    expected_path = os.path.join(model_pth_dir, "model_Ep5.pth")
    assert read_pth_from_model_loader(model_loader, model_pth_dir) == expected_path


def test_read_pth_from_model_loader_best():
    model_loader = {"load_way": "best"}
    model_pth_dir = "/path/to/models"
    expected_path = os.path.join(model_pth_dir, "best_model.pth")
    assert read_pth_from_model_loader(model_loader, model_pth_dir) == expected_path


def test_read_pth_from_model_loader_latest(mocker):
    model_loader = {"load_way": "latest"}
    model_pth_dir = "/path/to/models"
    latest_file = "latest_model.pth"
    mocker.patch(
        "torchhydro.trainers.train_utils.get_lastest_file_in_a_dir",
        return_value=latest_file,
    )
    expected_path = latest_file
    assert read_pth_from_model_loader(model_loader, model_pth_dir) == expected_path


def test_read_pth_from_model_loader_pth():
    model_loader = {"load_way": "pth", "pth_path": "/path/to/models/custom_model.pth"}
    model_pth_dir = "/path/to/models"
    expected_path = "/path/to/models/custom_model.pth"
    assert read_pth_from_model_loader(model_loader, model_pth_dir) == expected_path


def test_read_pth_from_model_loader_invalid():
    model_loader = {"load_way": "invalid"}
    model_pth_dir = "/path/to/models"
    with pytest.raises(ValueError, match="Invalid load_way"):
        read_pth_from_model_loader(model_loader, model_pth_dir)



import json
import os

# import matplotlib.pyplot as plt
import hydrodatasource.configs.config as hdscc
import numpy as np
import pandas as pd
import xarray as xr


def test_comp_compiler():
    origin_563_basins_era5l = "/ftproot/era5land/563_basins_era5_land.nc"
    origin_563_basins_gpm = "/ftproot/gpm_camels_hourly/563_basins_gpm.nc"
    smap_camels_sl_split_0 = "/ftproot/smap/camels_songliao_split_0.nc"
    origin_biliu_str = 's3://basins-origin/hour_data/1h/mean_data/streamflow_basin/streamflow_21401550.nc'
    final_forcing_biliu_era5l = 's3://basins-origin/hour_data/1h/mean_data/data_forcing_era5land/data_forcing_21401550.nc'
    final_forcing_biliu_gpm = 's3://basins-origin/hour_data/1h/mean_data/data_forcing_gpm/data_forcing_21401550.nc'
    final_forcing_str_biliu = 's3://basins-origin/hour_data/1h/mean_data/data_forcing_era5land_streamflow/data_forcing_streamflow_21401550.nc'
    split_era5land_result = split_basin_from_origin_era5land('era5land', origin_563_basins_era5l)
    split_gpm_result = split_basin_from_origin_gpm('gpm', origin_563_basins_gpm)
    split_smap_result = split_basin_from_smap(smap_camels_sl_split_0)
    streamflow_result = rename_origin_streamflow(origin_biliu_str)
    split_era5l_forcing_result = concat_forcing_smap('era5land', split_era5land_result, split_smap_result)
    split_gpm_forcing_result = concat_forcing_smap('gpm', split_gpm_result, split_smap_result)
    final_forcing_era5l_ds = xr.open_dataset(hdscc.FS.open(final_forcing_biliu_era5l))
    final_forcing_gpm_ds = xr.open_dataset(hdscc.FS.open(final_forcing_biliu_gpm))
    final_str_arr = xr.open_dataset(hdscc.FS.open(final_forcing_str_biliu))['streamflow']
    predct_str = streamflow_result.time >= np.datetime64('2015-03-31 01:00:00')
    predct_split_gpm = ((split_gpm_forcing_result.time >= np.datetime64('2015-03-31 16:00:00')) &
                        (split_gpm_forcing_result.time <= np.datetime64('2023-12-31 00:00:00')))
    predct_final_gpm = ((final_forcing_gpm_ds.time >= np.datetime64('2015-03-31 16:00:00')) &
                        (final_forcing_gpm_ds.time <= np.datetime64('2023-12-31 00:00:00')))
    split_gpm_forcing_result = split_gpm_forcing_result.sel(time=predct_split_gpm)
    final_forcing_gpm_ds = final_forcing_gpm_ds.sel(time=predct_final_gpm)
    print('______________________________')
    print(split_era5l_forcing_result)
    print('______________________________')
    print(split_gpm_forcing_result)
    print('______________________________')
    print(final_forcing_era5l_ds)
    print('______________________________')
    print(final_forcing_gpm_ds)
    for variable in final_forcing_era5l_ds.variables:
        print(np.array_equal(final_forcing_era5l_ds[variable].to_numpy(), split_era5l_forcing_result[variable].to_numpy()))
    print('______________________________')
    for variable in final_forcing_gpm_ds.variables:
        # ffgd = final_forcing_gpm_ds['gpm_tp'].to_numpy(), ffgd[69079]=0.045431413
        # sgfr = split_gpm_forcing_result['gpm_tp'].to_numpy(), sgfr[69079]=0.090862826
        print(np.array_equal(final_forcing_gpm_ds[variable].to_numpy(), split_gpm_forcing_result[variable].to_numpy()))
    print('______________________________')
    # print(streamflow_result['streamflow'])
    print(streamflow_result.sel(time=predct_str)['streamflow'])
    print('______________________________')
    print(final_str_arr)
    print(np.array_equal(final_str_arr.to_numpy(), streamflow_result.sel(time=predct_str)['streamflow'].to_numpy()))


def split_basin_from_origin_gpm(data_name, file_name):
    # copy from https://github.com/iHeadWater/HydroDataCompiler/blob/main/notebooks/convert_concat_origin_data_to_train_dataset.ipynb
    ds = xr.open_dataset(file_name)
    ds = ds.rename({'time_start': 'time', 'precipitationCal': 'gpm_tp'})
    ds = ds.drop_vars([var for var in ds.data_vars if var != 'gpm_tp'])
    ds['time'] = ds['time'].astype('datetime64[ns]')
    ds['basin_id'] = ds['basin_id'].astype(str).str.zfill(8)
    ds = ds.astype('float32')
    resampled_ds = ds.resample(time='h').mean()
    # for basin_id in ds['basin_id'].values:
    # only split out 1 basin from original data file
    basin_id = '21401550'
    basin_ds = resampled_ds.sel(basin_id=basin_id)
    basin_ds = basin_ds.drop('basin_id')
    output_dir = f'{data_name}_basin'
    os.makedirs(output_dir, exist_ok=True)
    output_file = f'{output_dir}/{data_name}_{basin_id}.nc'
    basin_ds.to_netcdf(output_file)
    return output_file


def split_basin_from_origin_era5land(data_name, file_name):
    # copy from https://github.com/iHeadWater/HydroDataCompiler/blob/main/notebooks/convert_concat_origin_data_to_train_dataset.ipynb
    ds = xr.open_dataset(file_name)
    ds = ds.rename({'time_start': 'time'})
    ds['time'] = ds['time'].astype('datetime64[ns]')
    ds = ds.astype('float32')
    ds['basin_id'] = ds['basin_id'].astype(str).str.zfill(8)
    # for basin_id in ds['basin_id'].values:
    # only split out 1 basin from original data file
    basin_id = '21401550'
    basin_ds = ds.sel(basin_id=basin_id)
    # basin_ds = basin_ds.drop('basin_id')
    output_dir = f'{data_name}_basin'
    os.makedirs(output_dir, exist_ok=True)
    output_file = f'{output_dir}/{data_name}_{basin_id}.nc'
    basin_ds.to_netcdf(output_file)
    return output_file


def split_basin_from_smap(file_path):
    # copy from https://github.com/iHeadWater/HydroDataCompiler/blob/main/notebooks/convert_concat_origin_data_to_train_dataset.ipynb
    ds = xr.open_dataset(file_path)
    ds = ds.rename({'time_start': 'time'})
    ds['time'] = ds['time'].astype('datetime64[ns]')
    ds['basin_id'] = ds['basin_id'].astype(str).str.zfill(8)
    ds = ds.astype('float32')
    ds = ds.drop_vars([var for var in ds.data_vars if var not in ['sm_surface', 'sm_rootzone']])
    data_shifted = ds.assign_coords(time=ds.time - pd.Timedelta(minutes=30))
    start_date = data_shifted['time'].values[0]
    end_date = data_shifted.time[-1].values
    hourly_time = pd.date_range(start=start_date, end=end_date, freq='h')
    data_hourly = data_shifted.reindex(time=hourly_time, method='ffill')
    # for basin_id in ds['basin_id'].values:
    # only split out 1 basin from original data file
    basin_id = '21401550'
    basin_ds = data_hourly.sel(basin_id=basin_id)
    basin_ds = basin_ds.drop('basin_id')
    output_file = f'smap_basin/smap_{basin_id}.nc'
    basin_ds.to_netcdf(output_file)
    return output_file


def concat_forcing_smap(data_name, file_path1, file_path2):
    basin_id = '21401550'
    if 's3' in file_path1:
        dataset1 = xr.open_dataset(hdscc.FS.open(file_path1))
    else:
        dataset1 = xr.open_dataset(file_path1)
    if 's3' in file_path2:
        dataset2 = xr.open_dataset(hdscc.FS.open(file_path1))
    else:
        dataset2 = xr.open_dataset(file_path2)
    dataset2_times = pd.to_datetime(dataset2.time.values)
    dataset1_times = pd.to_datetime(dataset1.time.values)
    common_times = pd.Index(dataset2_times).intersection(pd.Index(dataset1_times))
    dataset2_common = dataset2.sel(time=common_times)
    dataset1_common = dataset1.sel(time=common_times)
    combined_dataset = xr.merge([dataset2_common, dataset1_common])
    output_dir = f'data_forcing_{data_name}'
    os.makedirs(output_dir, exist_ok=True)
    combined_dataset.to_netcdf(f'{output_dir}/data_forcing_{basin_id}.nc')
    return combined_dataset


def rename_origin_streamflow(file_path):
    # copy from https://github.com/iHeadWater/HydroDataCompiler/blob/main/notebooks/convert_concat_origin_data_to_train_dataset.ipynb
    ds = xr.open_dataset(hdscc.FS.open(file_path))
    # ds = ds.rename({'TM': 'time', 'Q': 'streamflow'})
    ds = ds.astype('float32')
    basename = file_path.split('.')[-2]
    basin_id = basename.split('_')[-1]
    start_time = ds.time.min().values
    end_time = ds.time.max().values
    continuous_time_range = pd.date_range(start=start_time, end=end_time, freq='h')
    ds = ds.reindex(time=continuous_time_range)
    units_dict = {"streamflow": "mm/h"}
    units_json = json.dumps(units_dict)
    ds.attrs['units'] = units_json
    output_file = f'streamflow_basin/streamflow_{basin_id}.nc'
    ds.to_netcdf(output_file)
    return ds


'''
def plot_streamflow_arrays(sm):
    basin = '21401550'
    sm_surface = sm['streamflow']
    # 绘制曲线图
    plt.figure(figsize=(10, 5))
    plt.plot(sm_surface.time, sm_surface, label='sm_surface')
    plt.xlabel('Time')
    plt.ylabel('Streamflow')
    plt.title(f'{basin} Streamflow Over Time')
    plt.show()
'''



"""
Author: Xinzhuo Wu
Date: 2023-09-30 1:20:18
LastEditTime: 2024-05-27 16:26:06
LastEditors: Wenyu Ouyang
Description: spp lstm model
FilePath: \torchhydro\torchhydro\models\spplstm.py
Copyright (c) 2021-2022 Wenyu Ouyang. All rights reserved.
"""

import torch
from torch import nn


class TimeDistributed(nn.Module):
    def __init__(self, layer):
        super(TimeDistributed, self).__init__()
        self.layer = layer

    def forward(self, x):
        outputs = []
        for t in range(x.size(1)):
            xt = x[:, t, :]
            output = self.layer(xt)
            outputs.append(output.unsqueeze(1))
        outputs = torch.cat(outputs, dim=1)
        return outputs


class SppLayer(nn.Module):
    def __init__(self, out_pool_size):
        """
        out_pool_size: a int vector of expected output size of max pooling layer
        """
        super(SppLayer, self).__init__()
        self.out_pool_size = out_pool_size
        self.pools = []
        for i in range(len(out_pool_size)):
            pool_i = nn.AdaptiveMaxPool2d(out_pool_size[i])
            self.pools.append(pool_i)

    def forward(self, previous_conv):
        """
        Parameters
        ----------
        previous_conv
            a tensor vector of previous convolution layer

        Returns
        -------
        torch.Tensor
            a tensor vector with shape [1 x n] is the concentration of multi-level pooling
        """

        num_sample = previous_conv.size(0)
        channel_size = previous_conv.size(1)
        out_pool_size = self.out_pool_size
        for i in range(len(out_pool_size)):
            maxpool = self.pools[i]
            x = maxpool(previous_conv)
            if i == 0:
                spp = x.view(num_sample, channel_size, -1)
            else:
                spp = torch.cat((spp, x.view(num_sample, channel_size, -1)), -1)
        return spp


class SPP_LSTM_Model(nn.Module):
    def __init__(
        self, forecast_history, forecast_length, n_output, n_hidden_states, dropout
    ):
        super(SPP_LSTM_Model, self).__init__()

        self.conv1 = TimeDistributed(
            nn.Conv2d(
                in_channels=1,
                out_channels=32,
                kernel_size=1,
                padding="same",
            )
        )

        self.conv2 = TimeDistributed(
            nn.Conv2d(
                in_channels=32,
                out_channels=16,
                kernel_size=3,
                padding="same",
                bias=True,
            )
        )

        self.conv3 = TimeDistributed(
            nn.Conv2d(
                in_channels=16,
                out_channels=16,
                kernel_size=3,
                padding="same",
                bias=True,
            )
        )

        self.maxpool1 = TimeDistributed(nn.MaxPool2d(kernel_size=2, stride=(2, 2)))

        self.conv4 = TimeDistributed(
            nn.Conv2d(
                in_channels=16,
                out_channels=32,
                kernel_size=3,
                padding="same",
                bias=True,
            )
        )

        self.conv5 = TimeDistributed(
            nn.Conv2d(
                in_channels=32,
                out_channels=32,
                kernel_size=3,
                padding="same",
                bias=True,
            )
        )

        self.maxpool2 = TimeDistributed(SppLayer([4, 2, 1]))

        self.dropout = nn.Dropout(dropout)

        self.lstm = nn.LSTM(
            input_size=21 * 32, hidden_size=n_hidden_states, batch_first=True
        )

        self.dense = nn.Linear(in_features=n_hidden_states, out_features=n_output)

        self.forecast_length = forecast_length

    def forward(self, x):
        x = self.conv1(x)
        x = torch.relu(x)
        x = self.conv2(x)
        x = torch.relu(x)
        x = self.conv3(x)
        x = torch.relu(x)
        x = self.maxpool1(x)
        x = self.conv4(x)
        x = torch.relu(x)
        x = self.conv5(x)
        x = torch.relu(x)
        x = self.maxpool2(x)
        x = self.dropout(x)
        x = x.view(x.shape[0], x.shape[1], -1)
        x, _ = self.lstm(x)
        x = self.dense(x)
        x = x[:, -self.forecast_length :, :]
        return x


class SPP_LSTM_Model_2(nn.Module):
    def __init__(
        self,
        forecast_history,
        forecast_length,
        p_n_output,
        p_n_hidden_states,
        p_dropout,
        p_in_channels,
        p_out_channels,
        len_c=None,
        s_forecast_history=None,
        s_n_output=None,
        s_n_hidden_states=None,
        s_dropout=None,
        s_in_channels=None,
        s_out_channels=None,
    ):
        """
        A custom neural network model designed for handling and integrating various types of meteorological and geographical data.
        This model incorporates data from different sources: precipitation (p), soil (s), basin attributes (c), and Global Forecast System (g).

        Parameters:
        - forecast_history: The length of the input sequence for precipitation data.
        - forecast_length: The length of the forecast period.
        - p_n_output: Output dimension for the precipitation (p) data path.
        - p_n_hidden_states: Number of hidden states in the LSTM for the precipitation path.
        - p_dropout: Dropout rate applied in the precipitation path.
        - p_in_channels: Number of input channels for the convolutional layer in the precipitation path.
        - p_out_channels: Number of output channels for the convolutional layer in the precipitation path.
        - len_c: Optional, the number of basin attribute (c) features.
        - s_forecast_history, s_n_output, s_n_hidden_states, s_dropout, s_in_channels, s_out_channels: Similar parameters for the soil (s) data path.

        Methods:
        - forward(*x_lst): Processes the input data depending on its composition:
            - If only precipitation data is provided, processes it through the primary precipitation path.
            - If precipitation and either basin attributes or soil data are provided, integrates these with precipitation data for processing.
            - If all three types of data (precipitation, basin attributes, soil) are provided, combines them effectively for prediction.

        This model's architecture is versatile, enabling it to adapt to various data availability scenarios, making it suitable for complex meteorological and geographical data processing tasks.
        """
        super(SPP_LSTM_Model_2, self).__init__()
        self.conv_p = nn.Conv2d(
            in_channels=p_in_channels,
            out_channels=p_out_channels,
            kernel_size=(3, 3),
            padding="same",
        )

        self.leaky_relu_p = nn.LeakyReLU(0.01)

        self.lstm_p = nn.LSTM(
            input_size=p_out_channels * 5 + len_c,
            hidden_size=p_n_hidden_states,
            batch_first=True,
        )

        self.dropout_p = nn.Dropout(p_dropout)

        self.fc_p = nn.Linear(p_n_hidden_states, p_n_output)

        self.spp_p = SppLayer([2, 1])

        self.p_length = forecast_history + forecast_length
        self.forecast_length = forecast_length

        if s_forecast_history is not None:
            self.conv_s = nn.Conv2d(
                in_channels=s_in_channels,
                out_channels=s_out_channels,
                kernel_size=(3, 3),
                padding="same",
            )

            self.leaky_relu_s = nn.LeakyReLU(0.01)
            self.sigmoid_s = nn.Sigmoid()

            self.lstm_s = nn.LSTM(
                input_size=s_out_channels * 5,
                hidden_size=s_n_hidden_states,
                batch_first=True,
            )

            self.dropout_s = nn.Dropout(s_dropout)

            self.fc_s = nn.Linear(s_n_hidden_states, s_n_output)

            self.spp_s = SppLayer([2, 1])

            self.s_length = s_forecast_history

    def forward(self, *x_lst):
        # c and s must be None, g might be None
        if len(x_lst) == 1:
            x = x_lst[0]
            x = x.view(-1, x.shape[2], x.shape[3], x.shape[4])
            x = self.conv_p(x)
            x = self.leaky_relu_p(x)
            x = self.spp_p(x)
            x = x.view(x.shape[0], -1)
            x = x.view(int(x.shape[0] / (self.p_length)), self.p_length, -1)
            x, _ = self.lstm_p(x)
            x = self.dropout_p(x)
            x = self.fc_p(x)
        # g might be None. either c or s must be None, but not both
        elif len(x_lst) == 2:
            p = x_lst[0]
            m = x_lst[1].permute(1, 0, 2)
            # c is not None
            if m.dim() == 3:
                p = p.view(-1, p.shape[2], p.shape[3], p.shape[4])
                p = self.conv_p(p)
                p = self.leaky_relu_p(p)
                p = self.spp_p(p)
                p = p.view(p.shape[0], -1)
                p = p.view(int(p.shape[0] / (self.p_length)), self.p_length, -1)
                x = torch.cat([p, m], dim=2)
                x, _ = self.lstm_p(x)
                x = self.dropout_p(x)
                x = self.fc_p(x)
            # s is not None
            else:
                p = p.view(-1, p.shape[2], p.shape[3], p.shape[4])
                p = self.conv_p(p)
                p = self.leaky_relu_p(p)
                p = self.spp_p(p)
                p = p.view(p.shape[0], -1)
                p = p.view(int(p.shape[0] / (self.p_length)), self.p_length, -1)
                p, _ = self.lstm_p(p)
                p = self.dropout_p(p)
                p = self.fc_p(p)

                m = m.view(-1, m.shape[2], m.shape[3], m.shape[4])
                m = self.conv_s(m)
                m = self.leaky_relu_s(m)
                m = self.spp_s(m)
                m = m.view(m.shape[0], -1)
                m = m.view(int(m.shape[0] / (self.s_length)), self.s_length, -1)
                m, _ = self.lstm_s(m)
                m = m[:, -1:, :]
                m = self.dropout_s(m)
                m = self.fc_s(m)
                m = self.sigmoid_s(m)

                x = m * p
        # g might be None. Both s and c are not None
        elif len(x_lst) == 3:
            p = x_lst[0]
            c = x_lst[1].permute(1, 0, 2)
            s = x_lst[2]

            p = p.view(-1, p.shape[2], p.shape[3], p.shape[4])
            p = self.conv_p(p)
            p = self.leaky_relu_p(p)
            p = self.spp_p(p)
            p = p.view(p.shape[0], -1)
            p = p.view(int(p.shape[0] / (self.p_length)), self.p_length, -1)
            p_c = torch.cat([p, c], dim=2)
            p_c, _ = self.lstm_p(p_c)
            p_c = self.dropout_p(p_c)
            p_c = self.fc_p(p_c)

            s = s.view(-1, s.shape[2], s.shape[3], s.shape[4])
            s = self.conv_s(s)
            s = self.leaky_relu_s(s)
            s = self.spp_s(s)
            s = s.view(s.shape[0], -1)
            s = s.view(int(s.shape[0] / (self.s_length)), self.s_length, -1)
            s, _ = self.lstm_s(s)
            s = s[:, -1:, :]
            s = self.dropout_s(s)
            s = self.fc_s(s)
            s = self.sigmoid_s(s)

            x = s * p_c
        return x[:, -self.forecast_length :, :]
